import asyncio
import typing

from graphql import GraphQLError
from graphql.error import format_error as format_graphql_error
from starlette.requests import Request
from starlette.types import Receive, Scope, Send
from starlette.websockets import WebSocket, WebSocketState

from ..schema import BaseSchema
from ..utils.debug import pretty_print_graphql_operation
from .constants import (
    GQL_COMPLETE,
    GQL_CONNECTION_ACK,
    GQL_CONNECTION_INIT,
    GQL_CONNECTION_KEEP_ALIVE,
    GQL_CONNECTION_TERMINATE,
    GQL_DATA,
    GQL_START,
    GQL_STOP,
)
from .http import get_http_response


class GraphQL:
    def __init__(
        self,
        schema: BaseSchema,
        root_value: typing.Any = None,
        graphiql: bool = True,
        keep_alive: bool = False,
        keep_alive_interval: float = 1,
        debug: bool = False,
    ) -> None:
        self.schema = schema
        self.graphiql = graphiql
        self.root_value = root_value
        self.keep_alive = keep_alive
        self.keep_alive_interval = keep_alive_interval
        self._keep_alive_task = None
        self.debug = debug

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        if scope["type"] == "http":
            await self.handle_http(scope=scope, receive=receive, send=send)
        elif scope["type"] == "websocket":
            await self.handle_websocket(scope=scope, receive=receive, send=send)
        else:
            raise ValueError("Unknown scope type: %r" % (scope["type"],))

    async def handle_keep_alive(self, websocket):
        if websocket.application_state == WebSocketState.DISCONNECTED:
            return

        await websocket.send_json({"type": GQL_CONNECTION_KEEP_ALIVE})
        await asyncio.sleep(self.keep_alive_interval)

        self._keep_alive_task = asyncio.create_task(self.handle_keep_alive(websocket))

    async def handle_websocket(self, scope: Scope, receive: Receive, send: Send):
        websocket = WebSocket(scope=scope, receive=receive, send=send)

        await websocket.accept(subprotocol="graphql-ws")

        while websocket.application_state != WebSocketState.DISCONNECTED:
            message = await websocket.receive_json()

            operation_id = message.get("id")
            message_type = message.get("type")

            if message_type == GQL_CONNECTION_INIT:
                await websocket.send_json({"type": GQL_CONNECTION_ACK})

                if self.keep_alive:
                    asyncio.create_task(self.handle_keep_alive(websocket))
            elif message_type == GQL_CONNECTION_TERMINATE:
                await websocket.close()
            elif message_type == GQL_START:
                await self.start_subscription(
                    message.get("payload"), operation_id, websocket
                )
            elif message_type == GQL_STOP:
                await websocket.close()

    async def start_subscription(self, data, operation_id: str, websocket: WebSocket):
        query = data["query"]
        variables = data.get("variables")
        operation_name = data.get("operation_name")

        if self.debug:
            pretty_print_graphql_operation(operation_name, query, variables)

        context = {"websocket": websocket}

        data = await self.schema.subscribe(
            query,
            variable_values=variables,
            operation_name=operation_name,
            context_value=context,
        )

        try:
            async for result in data:
                payload = {"data": result.data}

                if result.errors:
                    payload["errors"] = [
                        format_graphql_error(err) for err in result.errors
                    ]
                await self._send_message(websocket, GQL_DATA, payload, operation_id)
        except Exception as error:
            if not isinstance(error, GraphQLError):
                error = GraphQLError(str(error), original_error=error)

            await self._send_message(
                websocket,
                GQL_DATA,
                {"data": None, "errors": [format_graphql_error(error)]},
                operation_id,
            )

        await self._send_message(websocket, GQL_COMPLETE, None, operation_id)

        if self._keep_alive_task:
            self._keep_alive_task.cancel()

        await websocket.close()

    async def _send_message(
        self,
        websocket: WebSocket,
        type_: str,
        payload: typing.Any = None,
        operation_id: str = None,
    ) -> None:
        data = {"type": type_}

        if operation_id is not None:
            data["id"] = operation_id

        if payload is not None:
            data["payload"] = payload

        return await websocket.send_json(data)

    async def handle_http(self, scope: Scope, receive: Receive, send: Send):
        request = Request(scope=scope, receive=receive)
        response = await get_http_response(request, self.execute, self.graphiql)

        await response(scope, receive, send)

    async def execute(self, query, variables=None, context=None, operation_name=None):
        if self.debug:
            pretty_print_graphql_operation(operation_name, query, variables)

        return await self.schema.execute(
            query,
            root_value=self.root_value,
            variable_values=variables,
            operation_name=operation_name,
            context_value=context,
        )



from dataclasses import InitVar, dataclass
from typing import Optional, Tuple, Type


@dataclass
class UnionDefinition:
    name: str
    description: Optional[str]
    types: InitVar[Tuple[Type]]

    def __post_init__(self, types):
        self._types = types

    @property  # type: ignore
    def types(self) -> Tuple[Type]:
        from .types.type_resolver import _resolve_generic_type

        types = tuple(
            _resolve_generic_type(t, self.name)
            for t in self._types
            if t is not None.__class__
        )

        return types  # type: ignore


def union(name: str, types: Tuple[Type], *, description=None):
    """Creates a new named Union type.

    Example usages:

    >>> strawberry.union(
    >>>     "Name",
    >>>     (A, B),
    >>> )

    >>> strawberry.union(
    >>>     "Name",
    >>>     (A, B),
    >>> )
    """

    union_definition = UnionDefinition(name=name, description=description, types=types)

    # This is currently a temporary solution, this is ok for now
    # But in future we might want to change this so that it works
    # properly with mypy, but there's no way to return a type like NewType does
    # so we return this class instance as it allows us to reuse the rest of
    # our code without doing too many changes

    def _call(self):
        raise ValueError("Cannot use union type directly")

    union_class = type(
        name, (), {"_union_definition": union_definition, "__call__": _call},
    )

    return union_class()



import dataclasses
from typing import Callable, List, Optional, Type, cast

from strawberry.exceptions import MissingReturnAnnotationError

from .arguments import get_arguments_from_resolver
from .permission import BasePermission
from .types.types import FederationFieldParams, FieldDefinition
from .utils.str_converters import to_camel_case


def check_return_annotation(field_definition: FieldDefinition):
    f = cast(Callable, field_definition.base_resolver)
    name = cast(str, field_definition.name)

    if "return" not in f.__annotations__:
        raise MissingReturnAnnotationError(name)


class StrawberryField(dataclasses.Field):
    _field_definition: FieldDefinition

    def __init__(self, field_definition: FieldDefinition):
        self._field_definition = field_definition

        super().__init__(  # type: ignore
            default=dataclasses.MISSING,
            default_factory=dataclasses.MISSING,
            init=field_definition.base_resolver is None,
            repr=True,
            hash=None,
            compare=True,
            metadata=None,
        )

    def __setattr__(self, name, value):
        if name == "type":
            self._field_definition.type = value

        if value and name == "name":
            if not self._field_definition.origin_name:
                self._field_definition.origin_name = value

            if not self._field_definition.name:
                self._field_definition.name = to_camel_case(value)

        return super().__setattr__(name, value)

    def __call__(self, f):
        f._field_definition = self._field_definition
        f._field_definition.name = f._field_definition.name or to_camel_case(f.__name__)
        f._field_definition.base_resolver = f
        f._field_definition.origin = f
        f._field_definition.arguments = get_arguments_from_resolver(
            f, f._field_definition.name
        )

        check_return_annotation(f._field_definition)

        f._field_definition.type = f.__annotations__["return"]

        return f


def field(
    f=None,
    *,
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    resolver: Optional[Callable] = None,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    federation: Optional[FederationFieldParams] = None
):
    """Annotates a method or property as a GraphQL field.

    This is normally used inside a type declaration:

    >>> @strawberry.type:
    >>> class X:
    >>>     field_abc: str = strawberry.field(description="ABC")

    >>>     @strawberry.field(description="ABC")
    >>>     def field_with_resolver(self, info) -> str:
    >>>         return "abc"

    it can be used both as decorator and as a normal function.
    """

    origin_name = f.__name__ if f else None
    name = name or (to_camel_case(origin_name) if origin_name else None)

    wrap = StrawberryField(
        field_definition=FieldDefinition(
            origin_name=origin_name,
            name=name,
            type=None,  # type: ignore
            origin=f,  # type: ignore
            description=description,
            base_resolver=resolver,
            is_subscription=is_subscription,
            permission_classes=permission_classes or [],
            arguments=(
                get_arguments_from_resolver(resolver, origin_name) if resolver else []
            ),
            federation=federation or FederationFieldParams(),
        )
    )

    if f:
        return wrap(f)

    return wrap



import dataclasses
from enum import EnumMeta
from typing import Any, List, Optional

from .exceptions import NotAnEnum


@dataclasses.dataclass
class EnumValue:
    name: str
    value: Any


@dataclasses.dataclass
class EnumDefinition:
    name: str
    values: List[EnumValue]
    description: Optional[str]


def _process_enum(cls, name=None, description=None):
    if not isinstance(cls, EnumMeta):
        raise NotAnEnum()

    if not name:
        name = cls.__name__

    description = description

    cls._enum_definition = EnumDefinition(
        name=name,
        values=[EnumValue(item.name, item.value) for item in cls],
        description=description,
    )

    return cls


def enum(_cls=None, *, name=None, description=None):
    """Registers the enum in the GraphQL type system.

    If name is passed, the name of the GraphQL type will be
    the value passed of name instead of the Enum class name.
    """

    def wrap(cls):
        return _process_enum(cls, name, description)

    if not _cls:
        return wrap

    return wrap(_cls)



import enum
from typing import Callable

from .arguments import convert_arguments
from .field import FieldDefinition
from .utils.inspect import get_func_args


def get_resolver(field: FieldDefinition) -> Callable:
    def _check_permissions(source, info, **kwargs):
        """
        Checks if the permission should be accepted and
        raises an exception if not
        """
        for permission_class in field.permission_classes:
            permission = permission_class()

            if not permission.has_permission(source, info, **kwargs):
                message = getattr(permission, "message", None)
                raise PermissionError(message)

    def _resolver(source, info, **kwargs):
        _check_permissions(source, info, **kwargs)

        actual_resolver = field.base_resolver

        if actual_resolver:
            kwargs = convert_arguments(kwargs, field.arguments)

            # the following code allows to omit info and root arguments
            # by inspecting the original resolver arguments,
            # if it asks for self, the source will be passed as first argument
            # if it asks for root, the source it will be passed as kwarg
            # if it asks for info, the info will be passed as kwarg

            function_args = get_func_args(actual_resolver)

            args = []

            if "self" in function_args:
                args.append(source)

            if "root" in function_args:
                kwargs["root"] = source

            if "info" in function_args:
                kwargs["info"] = info

            result = actual_resolver(*args, **kwargs)
        else:
            result = getattr(source, field.origin_name)

        # graphql-core expects a resolver for an Enum type to return
        # the enum's *value* (not its name or an instance of the enum).
        if isinstance(result, enum.Enum):
            return result.value

        return result

    return _resolver



import typing
from inspect import isawaitable

from graphql import (
    ExecutionResult,
    GraphQLError,
    GraphQLSchema,
    execute as graphql_execute,
    parse,
)
from graphql.subscription import subscribe as graphql_subscribe
from graphql.type import validate_schema
from graphql.validation import validate

from .middleware import Middleware


async def execute(
    schema: GraphQLSchema,
    query: str,
    root_value: typing.Any = None,
    context_value: typing.Any = None,
    variable_values: typing.Dict[str, typing.Any] = None,
    middleware: typing.List[Middleware] = None,
    operation_name: str = None,
):  # pragma: no cover
    schema_validation_errors = validate_schema(schema)
    if schema_validation_errors:
        return ExecutionResult(data=None, errors=schema_validation_errors)

    try:
        document = parse(query)
    except GraphQLError as error:
        return ExecutionResult(data=None, errors=[error])
    except Exception as error:
        error = GraphQLError(str(error), original_error=error)
        return ExecutionResult(data=None, errors=[error])

    validation_errors = validate(schema, document)

    if validation_errors:
        return ExecutionResult(data=None, errors=validation_errors)

    result = graphql_execute(
        schema,
        parse(query),
        root_value=root_value,
        middleware=middleware or [],
        variable_values=variable_values,
        operation_name=operation_name,
        context_value=context_value,
    )
    if isawaitable(result):
        result = await typing.cast(typing.Awaitable[ExecutionResult], result)
    return result


async def subscribe(
    schema: GraphQLSchema,
    query: str,
    root_value: typing.Any = None,
    context_value: typing.Any = None,
    variable_values: typing.Dict[str, typing.Any] = None,
    operation_name: str = None,
) -> typing.Union[
    typing.AsyncIterator[ExecutionResult], ExecutionResult
]:  # pragma: no cover
    document = parse(query)

    return await graphql_subscribe(
        schema=schema,
        document=document,
        root_value=root_value,
        context_value=context_value,
        variable_values=variable_values,
        operation_name=operation_name,
    )



import inspect
import typing
from itertools import islice

from graphql import GraphQLDirective
from strawberry.arguments import ArgumentDefinition, get_arguments_from_annotations

from .arguments import convert_arguments
from .types import TypeMap


def get_arguments_for_directive(
    resolver: typing.Callable,
) -> typing.List[ArgumentDefinition]:
    # TODO: move this into directive declaration
    annotations = resolver.__annotations__
    annotations = dict(islice(annotations.items(), 1, None))
    annotations.pop("return", None)

    parameters = inspect.signature(resolver).parameters

    return get_arguments_from_annotations(annotations, parameters, origin=resolver)


def get_directive_type(origin: typing.Any, type_map: TypeMap) -> GraphQLDirective:
    if not hasattr(origin, "directive_definition"):
        raise ValueError(f"Wrong type passed to get directive type {origin}")

    return GraphQLDirective(
        name=origin.directive_definition.name,
        locations=origin.directive_definition.locations,
        description=origin.directive_definition.description,
        args=convert_arguments(
            get_arguments_for_directive(origin.directive_definition.resolver), type_map
        ),
    )



import typing

from graphql import GraphQLField, GraphQLInputField
from strawberry.field import FieldDefinition
from strawberry.resolvers import get_resolver

from .arguments import convert_arguments
from .type import get_graphql_type
from .types import Field, TypeMap


def get_field(field: FieldDefinition, is_input: bool, type_map: TypeMap,) -> Field:
    graphql_type = get_graphql_type(field, type_map)

    TypeClass: typing.Union[
        typing.Type[GraphQLInputField], typing.Type[GraphQLField]
    ] = GraphQLField

    kwargs: typing.Dict[str, typing.Any] = {
        "description": field.description,
    }

    resolver = get_resolver(field)

    if is_input:
        TypeClass = GraphQLInputField
    elif field.is_subscription:
        kwargs["args"] = convert_arguments(field.arguments, type_map)
        kwargs["subscribe"] = resolver
        kwargs["resolve"] = lambda event, *args, **kwargs: event
    else:
        kwargs["args"] = convert_arguments(field.arguments, type_map)
        kwargs["resolve"] = resolver

    return TypeClass(graphql_type, **kwargs)  # type: ignore



import json
from typing import Any, Optional

from flask import Response, abort, render_template_string, request
from flask.views import View
from graphql.error import format_error as format_graphql_error

from ..schema import BaseSchema
from .graphiql import render_graphiql_page


class GraphQLView(View):
    methods = ["GET", "POST"]

    def __init__(
        self,
        schema: BaseSchema,
        graphiql: bool = True,
        root_value: Optional[Any] = None,
    ):
        self.graphiql = graphiql
        self.schema = schema
        self.root_value = root_value

    def render_template(self, request, template=None):
        return render_template_string(template)

    def dispatch_request(self):
        if "text/html" in request.environ.get("HTTP_ACCEPT", ""):
            if not self.graphiql:
                abort(404)

            template = render_graphiql_page()
            return self.render_template(request, template=template)

        data = request.json

        try:
            query = data["query"]
            variables = data.get("variables")
            operation_name = data.get("operationName")

        except KeyError:
            return Response("No valid query was provided for the request", 400)

        context = {"request": request}

        result = self.schema.execute_sync(
            query,
            variable_values=variables,
            context_value=context,
            operation_name=operation_name,
            root_value=self.root_value,
        )

        response_data = {"data": result.data}

        if result.errors:
            response_data["errors"] = [
                format_graphql_error(err) for err in result.errors
            ]

        return Response(
            json.dumps(response_data),
            status=400 if result.errors else 200,
            content_type="application/json",
        )



import dataclasses
from functools import partial
from typing import List, Optional, Type, cast

from strawberry.utils.typing import is_generic

from .types.types import FederationTypeParams, TypeDefinition
from .utils.str_converters import to_camel_case


def _get_interfaces(cls: Type) -> List[TypeDefinition]:
    interfaces = []

    for base in cls.__bases__:
        type_definition = cast(
            Optional[TypeDefinition], getattr(base, "_type_definition", None)
        )

        if type_definition and type_definition.is_interface:
            interfaces.append(type_definition)

    return interfaces


def _process_type(
    cls,
    *,
    name: Optional[str] = None,
    is_input: bool = False,
    is_interface: bool = False,
    description: Optional[str] = None,
    federation: Optional[FederationTypeParams] = None
):
    name = name or to_camel_case(cls.__name__)

    wrapped = dataclasses.dataclass(cls)

    interfaces = _get_interfaces(wrapped)

    wrapped._type_definition = TypeDefinition(
        name=name,
        is_input=is_input,
        is_interface=is_interface,
        is_generic=is_generic(cls),
        interfaces=interfaces,
        description=description,
        federation=federation or FederationTypeParams(),
        origin=cls,
    )

    return wrapped


def type(
    cls: Type = None,
    *,
    name: str = None,
    is_input: bool = False,
    is_interface: bool = False,
    description: str = None,
    federation: Optional[FederationTypeParams] = None
):
    """Annotates a class as a GraphQL type.

    Example usage:

    >>> @strawberry.type:
    >>> class X:
    >>>     field_abc: str = "ABC"
    """

    def wrap(cls):
        return _process_type(
            cls,
            name=name,
            is_input=is_input,
            is_interface=is_interface,
            description=description,
            federation=federation,
        )

    if cls is None:
        return wrap

    return wrap(cls)


input = partial(type, is_input=True)
interface = partial(type, is_interface=True)



import dataclasses
import sys
from typing import Dict, List, Optional, Type, Union, cast

from strawberry.exceptions import MissingTypesForGenericError
from strawberry.lazy_type import LazyType
from strawberry.union import union
from strawberry.utils.str_converters import to_camel_case
from strawberry.utils.typing import (
    get_args,
    get_async_generator_annotation,
    get_list_annotation,
    get_optional_annotation,
    get_parameters,
    has_type_var,
    is_async_generator,
    is_forward_ref,
    is_list,
    is_optional,
    is_type_var,
    is_union,
)

from .generics import copy_type_with, get_name_from_types
from .types import ArgumentDefinition, FieldDefinition


def _resolve_generic_type(type: Type, field_name: str) -> Type:
    if hasattr(type, "_type_definition") and type._type_definition.is_generic:
        args = get_args(type)

        # raise an error when using generics without passing any type parameter, ie:
        # >>> class X(Generic[T]): ...
        # >>> a: X
        # instead of
        # >>> a: X[str]

        if len(args) == 0:
            raise MissingTypesForGenericError(field_name, type)

        # we only make a copy when all the arguments are not type vars
        if not all(is_type_var(a) for a in args):
            return copy_type_with(type, *args)

    return type


def resolve_type(field_definition: Union[FieldDefinition, ArgumentDefinition]) -> None:
    # convert a python type to include a strawberry definition, so for example
    # Union becomes a class with a UnionDefinition, Generics become an actual
    # type definition. This helps with making the code to convert the type definitions
    # to GraphQL types, as we only have to deal with Python's typings in one place.

    type = cast(Type, field_definition.type)
    origin_name = cast(str, field_definition.origin_name)

    if isinstance(type, LazyType):
        field_definition.type = type.resolve_type()

    if isinstance(type, str):
        module = sys.modules[field_definition.origin.__module__].__dict__

        type = eval(type, module)
        field_definition.type = type

    if is_forward_ref(type):
        # if the type is a forward reference we try to resolve the type by
        # finding it in the global namespace of the module where the field
        # was intially declared. This will break when the type is not declared
        # in the main scope, but we don't want to support that use case
        # see https://mail.python.org/archives/list/typing-sig@python.org/thread/SNKJB2U5S74TWGDWVD6FMXOP63WVIGDR/  # noqa: E501

        type_name = type.__forward_arg__

        module = sys.modules[field_definition.origin.__module__]

        # TODO: we should probably raise an error if we can't find the type
        type = module.__dict__[type_name]

        field_definition.type = type

        return

    if is_async_generator(type):
        # TODO: shall we raise a warning if field is not used in a subscription?

        # async generators are used in subscription, we only need the yield type
        # https://docs.python.org/3/library/typing.html#typing.AsyncGenerator
        field_definition.type = get_async_generator_annotation(type)

        return resolve_type(field_definition)

    # check for Optional[A] which is represented as Union[A, None], we
    # have an additional check for proper unions below
    if is_optional(type) and len(type.__args__) == 2:
        # this logics works around List of optionals and Optional lists of Optionals:
        # >>> Optional[List[Str]]
        # >>> Optional[List[Optional[Str]]]
        # the field is only optional if it is not a list or if it was already optional
        # since we mark the child as optional when the field is a list

        field_definition.is_optional = (
            True and not field_definition.is_list or field_definition.is_optional
        )
        field_definition.is_child_optional = field_definition.is_list
        field_definition.type = get_optional_annotation(type)

        return resolve_type(field_definition)

    elif is_list(type):
        # TODO: maybe this should be an argument definition when it is argument
        # but doesn't matter much
        child_definition = FieldDefinition(
            origin=field_definition.origin,  # type: ignore
            name=None,
            origin_name=None,
            type=get_list_annotation(type),
        )

        resolve_type(child_definition)

        field_definition.type = None
        field_definition.is_list = True
        field_definition.child = child_definition

        return

    # case for Union[A, B, C], it also handles Optional[Union[A, B, C]] as optionals
    # type hints are represented as Union[..., None].

    elif is_union(type):
        # Optional[Union[A, B]] is represented as Union[A, B, None] so we need
        # too check again if the field is optional as the check above only checks
        # for single Optionals
        field_definition.is_optional = is_optional(type)

        types = type.__args__

        # we use a simplified version of resolve_type since unions in GraphQL
        # are simpler and cannot contain lists or optionals

        types = tuple(
            _resolve_generic_type(t, origin_name)
            for t in types
            if t is not None.__class__
        )

        field_definition.is_union = True
        field_definition.type = union(get_name_from_types(types), types)

    # case for Type[A], we want to convert generics to have the concrete types
    # when we pass them, so that we don't have to deal with generics when
    # generating the GraphQL types later on.

    elif hasattr(type, "_type_definition") and type._type_definition.is_generic:
        args = get_args(type)

        # raise an error when using generics without passing any type parameter, ie:
        # >>> class X(Generic[T]): ...
        # >>> a: X
        # instead of
        # >>> a: X[str]

        if len(args) == 0:
            name = cast(str, field_definition.origin_name)

            raise MissingTypesForGenericError(name, type)

        # we only make a copy when all the arguments are not type vars
        if not all(is_type_var(a) for a in args):
            field_definition.type = copy_type_with(type, *args)

    if hasattr(type, "_union_definition"):
        field_definition.is_union = True


def _get_type_params_for_field(
    field_definition: FieldDefinition,
) -> Optional[List[Type]]:
    if field_definition.is_list:
        child = cast(FieldDefinition, field_definition.child)

        return _get_type_params_for_field(child)

    type = cast(Type, field_definition.type)

    if hasattr(type, "_union_definition"):
        types = type._union_definition.types
        type_vars = [t for t in types if is_type_var(t)]

        if type_vars:
            return type_vars

    if is_type_var(type):
        return [type]

    if has_type_var(type):
        return get_parameters(type)

    return None


def _get_type_params(fields: List[FieldDefinition]) -> Dict[str, Type]:
    type_params = {}

    for field in fields:
        name = cast(str, field.origin_name)
        params = _get_type_params_for_field(field)

        # TODO: support multiple
        if params:
            type_params[name] = params[0]

    return type_params


def _resolve_types(fields: List[FieldDefinition]) -> List[FieldDefinition]:
    for field in fields:
        resolve_type(field)

    return fields


def _get_fields(cls: Type) -> List[FieldDefinition]:
    fields = []

    # get all the fields from the dataclass
    dataclass_fields = dataclasses.fields(cls)

    # plus the fields that are defined with the resolvers, using
    # the @strawberry.field decorator
    dataclass_fields += tuple(
        field for field in cls.__dict__.values() if hasattr(field, "_field_definition")
    )

    seen_fields = set()

    for field in dataclass_fields:
        if hasattr(field, "_field_definition"):
            field_definition = field._field_definition  # type: ignore

            # we make sure that the origin is either the field's resolver
            # when called as:
            # >>> @strawberry.field
            # >>> def x(self): ...
            # or the class where this field was defined, so we always have
            # the correct origin for determining field types when resolving
            # the types.

            field_definition.origin = field_definition.origin or cls
        else:
            # for fields that don't have a field definition, we create one
            # based on the dataclass field

            field_definition = FieldDefinition(
                origin_name=field.name,
                name=to_camel_case(field.name),
                type=field.type,
                origin=cls,
            )

        fields.append(field_definition)
        seen_fields.add(field_definition.origin_name)

    # let's also add fields that are declared with @strawberry.field in
    # parent classes, we do this by checking if parents have a type definition
    # and we haven't seen a field already

    # TODO: maybe we want to add a warning when overriding a field, as it might be
    # a mistake

    for base in cls.__bases__:
        if hasattr(base, "_type_definition"):
            fields += [
                field
                for field in base._type_definition.fields  # type: ignore
                if field.origin_name not in seen_fields
            ]

    return fields



from typing import Callable, Optional

from mypy.plugin import AnalyzeTypeContext, ClassDefContext, Plugin
from mypy.plugins import dataclasses
from mypy.types import Type


def lazy_type_analyze_callback(ctx: AnalyzeTypeContext) -> Type:
    type_name = ctx.type.args[0]
    type_ = ctx.api.analyze_type(type_name)

    return type_


class StrawberryPlugin(Plugin):
    def get_type_analyze_hook(self, fullname: str):
        if fullname == "strawberry.lazy_type.LazyType":
            return lazy_type_analyze_callback

        return None

    def get_class_decorator_hook(
        self, fullname: str
    ) -> Optional[Callable[[ClassDefContext], None]]:
        if any(
            strawberry_decorator in fullname
            for strawberry_decorator in {
                "strawberry.type",
                "strawberry.federation.type",
                "strawberry.input",
                "strawberry.interface",
            }
        ):
            return dataclasses.dataclass_class_maker_callback
        return None


def plugin(version: str):
    return StrawberryPlugin



import json
import os
import typing

from django.http import Http404, HttpResponseNotAllowed, JsonResponse
from django.http.response import HttpResponseBadRequest
from django.template import RequestContext, Template
from django.template.exceptions import TemplateDoesNotExist
from django.template.loader import render_to_string
from django.template.response import TemplateResponse
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

import strawberry
from graphql.error import format_error as format_graphql_error

from ..schema import BaseSchema


class GraphQLView(View):
    graphiql = True
    schema: typing.Optional[BaseSchema] = None

    def __init__(self, schema: BaseSchema, graphiql=True):
        self.schema = schema
        self.graphiql = graphiql

    def get_root_value(self):
        return None

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        if request.method.lower() not in ("get", "post"):
            return HttpResponseNotAllowed(
                ["GET", "POST"], "GraphQL only supports GET and POST requests."
            )

        if "text/html" in request.META.get("HTTP_ACCEPT", ""):
            if not self.graphiql:
                raise Http404("GraphiQL has been disabled")

            return self._render_graphiql(request)

        data = json.loads(request.body)

        try:
            query = data["query"]
            variables = data.get("variables")
            operation_name = data.get("operationName")
        except KeyError:
            return HttpResponseBadRequest("No GraphQL query found in the request")

        context = {"request": request}

        result = self.schema.execute_sync(
            query,
            root_value=self.get_root_value(),
            variable_values=variables,
            context_value=context,
            operation_name=operation_name,
        )

        response_data = {"data": result.data}

        if result.errors:
            response_data["errors"] = [
                format_graphql_error(err) for err in result.errors
            ]

        return JsonResponse(response_data)

    def _render_graphiql(self, request, context=None):
        try:
            template = Template(render_to_string("graphql/graphiql.html"))
        except TemplateDoesNotExist:
            template = Template(
                open(
                    os.path.join(
                        os.path.dirname(os.path.abspath(strawberry.__file__)),
                        "static/graphiql.html",
                    ),
                    "r",
                ).read()
            )

        context = context or {}
        context.update({"SUBSCRIPTION_ENABLED": "false"})

        response = TemplateResponse(request=request, template=None, context=context)
        response.content = template.render(RequestContext(request, context))

        return response



# TODO: add links to docs

from typing import List, Set

from graphql import GraphQLObjectType


class NotAnEnum(Exception):
    def __init__(self):
        message = "strawberry.enum can only be used with subclasses of Enum"

        super().__init__(message)


class MissingReturnAnnotationError(Exception):
    """The field is missing the return annotation"""

    def __init__(self, field_name: str):
        message = (
            f'Return annotation missing for field "{field_name}", '
            "did you forget to add it?"
        )

        super().__init__(message)


class MissingArgumentsAnnotationsError(Exception):
    """The field is missing the annotation for one or more arguments"""

    def __init__(self, field_name: str, arguments: Set[str]):
        arguments_list: List[str] = sorted(list(arguments))

        if len(arguments_list) == 1:
            argument = f'argument "{arguments_list[0]}"'
        else:
            head = ", ".join(arguments_list[:-1])
            argument = f'arguments "{head}" and "{arguments_list[-1]}"'

        message = (
            f"Missing annotation for {argument} "
            f'in field "{field_name}", did you forget to add it?'
        )

        super().__init__(message)


class WrongReturnTypeForUnion(Exception):
    """The Union type cannot be resolved because it's not a field"""

    def __init__(self, field_name: str, result_type: str):
        message = (
            f'The type "{result_type}" cannot be resolved for the field "{field_name}" '
            ", are you using a strawberry.field?"
        )

        super().__init__(message)


class UnallowedReturnTypeForUnion(Exception):
    """The return type is not in the list of Union types"""

    def __init__(
        self, field_name: str, result_type: str, allowed_types: Set[GraphQLObjectType]
    ):
        formatted_allowed_types = [type_.name for type_ in allowed_types]

        message = (
            f'The type "{result_type}" of the field "{field_name}" '
            f'is not in the list of the types of the union: "{formatted_allowed_types}"'
        )

        super().__init__(message)


class MissingTypesForGenericError(Exception):
    """Raised when a generic types was used without passing any type."""

    def __init__(self, field_name: str, annotation):
        message = (
            f'The type "{annotation.__name__}" of the field "{field_name}" '
            f"is generic, but no type has been passed"
        )

        super().__init__(message)


class UnsupportedTypeError(Exception):
    def __init__(self, annotation):
        message = f"{annotation} conversion is not supported"

        super().__init__(message)



from typing import Type, cast

from graphql import GraphQLInputObjectType, GraphQLObjectType
from graphql.type.definition import GraphQLInterfaceType
from strawberry.type import TypeDefinition

from .fields import get_field
from .types import ConcreteType, GraphQLType, TypeMap


def _get_object_type_for_type_definition(
    type_definition: TypeDefinition, type_map: TypeMap
) -> GraphQLType:

    TypeClass: Type = GraphQLObjectType

    kwargs = {}

    if type_definition.is_input:
        TypeClass = GraphQLInputObjectType
    elif type_definition.is_interface:
        TypeClass = GraphQLInterfaceType

    if type_definition.interfaces:
        kwargs["interfaces"] = [
            _get_object_type_for_type_definition(interface, type_map)
            for interface in type_definition.interfaces
        ]

    assert not type_definition.is_generic

    return TypeClass(
        name=type_definition.name,
        fields=lambda: {
            field.name: get_field(field, type_definition.is_input, type_map)
            for field in type_definition.fields
        },
        description=type_definition.description,
        **kwargs,
    )


def get_object_type(origin: Type, type_map: TypeMap) -> GraphQLObjectType:
    """Returns a root type (Query, Mutation, Subscription) from a decorated type"""

    if not hasattr(origin, "_type_definition"):
        raise ValueError(f"Wrong type passed to get object type {origin}")

    type_definition: TypeDefinition = origin._type_definition

    name = type_definition.name

    if name not in type_map:
        object_type = _get_object_type_for_type_definition(type_definition, type_map)

        type_map[name] = ConcreteType(
            definition=type_definition, implementation=object_type
        )

    return cast(GraphQLObjectType, type_map[name].implementation)



__version__ = "0.1.0"


from . import federation  # noqa
from .custom_scalar import scalar  # noqa
from .directive import directive  # noqa
from .enum import enum  # noqa
from .field import field  # noqa
from .lazy_type import LazyType  # noqa
from .mutation import mutation, subscription  # noqa
from .permission import BasePermission  # noqa
from .scalars import ID  # noqa
from .schema import Schema  # noqa
from .type import input, interface, type  # noqa
from .union import union  # noqa



from typing import cast

from graphql import GraphQLEnumType, GraphQLEnumValue
from strawberry.enum import EnumDefinition

from .types import ConcreteType, TypeMap


def get_enum_type(
    enum_definition: EnumDefinition, type_map: TypeMap
) -> GraphQLEnumType:
    if enum_definition.name not in type_map:
        enum = GraphQLEnumType(
            name=enum_definition.name,
            values={
                item.name: GraphQLEnumValue(item.value)
                for item in enum_definition.values
            },
            description=enum_definition.description,
        )

        type_map[enum_definition.name] = ConcreteType(
            definition=enum_definition, implementation=enum
        )

    return cast(GraphQLEnumType, type_map[enum_definition.name].implementation)



from typing import Any, Dict, List, Optional, Type, Union

from graphql import GraphQLSchema, graphql_sync, parse
from graphql.subscription import subscribe
from graphql.type.directives import specified_directives
from strawberry.custom_scalar import ScalarDefinition
from strawberry.enum import EnumDefinition
from strawberry.types.types import TypeDefinition

# TODO: get rid of this module ?
from ..graphql import execute
from ..middleware import DirectivesMiddleware, Middleware
from ..printer import print_schema
from .types import ConcreteType, get_directive_type, get_object_type


class Schema:
    def __init__(
        self,
        # TODO: can we make sure we only allow to pass something that has been decorated?
        query: Type,
        mutation: Optional[Type] = None,
        subscription: Optional[Type] = None,
        directives=(),
        types=(),
    ):

        self.type_map: Dict[str, ConcreteType] = {}

        query_type = get_object_type(query, self.type_map)
        mutation_type = get_object_type(mutation, self.type_map) if mutation else None
        subscription_type = (
            get_object_type(subscription, self.type_map) if subscription else None
        )

        self.middleware: List[Middleware] = [DirectivesMiddleware(directives)]

        directives = [
            get_directive_type(directive, self.type_map) for directive in directives
        ]

        self._schema = GraphQLSchema(
            query=query_type,
            mutation=mutation_type,
            subscription=subscription_type if subscription else None,
            directives=specified_directives + directives,
            types=[get_object_type(type, self.type_map) for type in types],
        )

        self.query = self.type_map[query_type.name]

    def get_type_by_name(
        self, name: str
    ) -> Optional[Union[TypeDefinition, ScalarDefinition, EnumDefinition]]:
        if name in self.type_map:
            return self.type_map[name].definition

        return None

    # TODO: type return value of these

    async def execute(
        self,
        query: str,
        variable_values: Optional[Dict[str, Any]] = None,
        context_value: Optional[Any] = None,
        root_value: Optional[Any] = None,
        operation_name: Optional[str] = None,
    ):
        return await execute(
            self._schema,
            query,
            variable_values=variable_values,
            root_value=root_value,
            context_value=context_value,
            middleware=self.middleware,
            operation_name=operation_name,
        )

    def execute_sync(
        self,
        query: str,
        variable_values: Optional[Dict[str, Any]] = None,
        context_value: Optional[Any] = None,
        root_value: Optional[Any] = None,
        operation_name: Optional[str] = None,
    ):
        return graphql_sync(
            self._schema,
            query,
            variable_values=variable_values,
            root_value=root_value,
            context_value=context_value,
            middleware=self.middleware,
            operation_name=operation_name,
        )

    async def subscribe(
        self,
        query: str,
        variable_values: Optional[Dict[str, Any]] = None,
        context_value: Optional[Any] = None,
        root_value: Optional[Any] = None,
        operation_name: Optional[str] = None,
    ):
        return await subscribe(
            self._schema,
            parse(query),
            root_value=root_value,
            context_value=context_value,
            variable_values=variable_values,
            operation_name=operation_name,
        )

    def as_str(self) -> str:
        return print_schema(self)
